/*=========================================================================

  Copyright Brigham and Women's Hospital (BWH) All Rights Reserved.

  See Doc/copyright/copyright.txt
  or http://www.slicer.org/copyright/copyright.txt for details.

  Program:   vtkvmtkITK

==========================================================================*/

/**
 * @file
 * @brief Common include pulling in the Utilities/vtkvmtkITK module's
 * Windows export-macro header; included by every header in this module.
 */

#include "vtkvmtkWin32Header.h"
