/* aggregate Stellar sources in the proper order */
#include "top.c"
#include "interact.c"
#include "vector.c"
#include "anisotropy.c"
#include "quality.c"
#include "arraypoolstack.c"
#include "journal.c"
#include "print.c"
#include "classify.c"
#include "quadric.c"
#include "smoothing.c"
#include "topological.c"
#include "output.c"
#include "insertion.c"
#include "size.c"
#include "improve.c"
#include "main.c"

