/*=========================================================================

Program:   VMTK 
Module:    vtkvmtkInteractorStyleTrackballCamera
Language:  C++
Date:      $Date: 2006/07/17 09:52:56 $
Version:   $Revision: 1.1 $

  Copyright (c) Luca Antiga, David Steinman. All rights reserved.
  See LICENSE file for details.

  Portions of this code are covered under the VTK copyright.
  See VTKCopyright.txt or http://www.kitware.com/VTKCopyright.htm 
  for details.

     This software is distributed WITHOUT ANY WARRANTY; without even 
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#include "vtkvmtkInteractorStyleTrackballCamera.h"
#include "vtkObjectFactory.h"


vtkStandardNewMacro(vtkvmtkInteractorStyleTrackballCamera);

void vtkvmtkInteractorStyleTrackballCamera::OnChar()
{
}

